# -*- coding: utf-8 -*-
"""
    A set of functions to work with W7-X Archivedb via web-api.

    Simple function to read from and to write to the database are available.
    This includes reading/writing arrays, images. Also reading of
    performed programs is available.

    The original implementation is by Dinesh Kanan. Small corrections
    and modifications are by S.A. Bozhenkov.
"""
from __future__ import print_function
from builtins import int
try:
    import urllib2 #Python 2
except ImportError:
    import urllib.request as urllib2 #Python 3
import json
import datetime,calendar
import numpy as np
import archivedb
from io import BytesIO
from PIL import Image
import os
import tables,random
import collections
import socket
import time
try:
    import configparser #Python 3
except ImportError:
    import ConfigParser as configparser #Python 2

__version__ = u"0.2.3"
__author__ = u"dsk"
__cache_path__ = os.path.expanduser(u"~/.archivedb/cache/")
__db_url__ = "http://archive-webapi.ipp-hgw.mpg.de/"
__mdsplus_query_="http://10.44.4.11/operator/find_pulse/"
__t02id__ = {}
__id2t0__ = {}
__id2fromto__ = {}

def get_json(url, timeout = 1, retries=10):
    request = urllib2.Request(url, headers={"Accept" : "application/json"})
    for i in range(retries):
        try:
            response = urllib2.urlopen(request, timeout=timeout)
            d = json.loads(response.read().decode("utf-8"))
            return d
        except socket.timeout:
            continue
    raise RuntimeError("failed to urlopen")

def print_info():
    """
        Print version, author, etc.
    """
    print(__doc__)
    print("Version: {}".format(__version__))
    print("Author: {}".format(__author__))
    print("Cache path: {}".format(__cache_path__))
    print("DB URL: {}".format(__db_url__))
    print("MDSPlus: {}".format(__mdsplus_query_))

def to_timestamp(time_from_user, fmt=u"%Y-%m-%d %H:%M:%S"):    
    """    
        Converts a date time string to the UTC time stamp in nanoseconds.
        
        Parameters
        ----------
        time_from_user : string       
            Of the form 'YYYY-MM-DD HH:MM:SS.f' or 'YYYY-MM-DD HH:MM:SS'

        Returns
        -------
        out : int
            The corresponding UTC time stamp in nanoseconds.  
    """
    if(len(time_from_user)>19):
        temp = time_from_user[0:19]
        time_dateobj = datetime.datetime.strptime(temp, fmt)
        
        extranano = int(float(time_from_user[19:len(time_from_user)])*1000000000)
    else:
        time_dateobj = datetime.datetime.strptime(time_from_user, fmt)
        extranano = 0
    
    time_nano = (int(round(calendar.timegm(time_dateobj.timetuple()) * 1000) 
                         +time_dateobj.microsecond)*1000000)+extranano
    return int(time_nano)

def to_stringdate(time_nano):    
    """    
        Converts a given UTC timestamp to a date strin.
        
        Parameters
        ----------
        time_nano : int
            The time in UTC nanoseconds.
            Of the form: 1435601915999999999
        Returns
        -------
        to_timestamp : str
    """
    mytime = datetime.datetime.utcfromtimestamp(time_nano/1e9)
    return mytime.strftime('%Y-%m-%d %H:%M:%S.%f')

def get_program_list(t0, t1):
    """Get list of performed experimental programs.

       This function returns a list of dictionaries for programs
       performed between start and stop times. The times are to be
       specified as '2016-03-10 12:00:00', or as ns timestamps.
    Parameters
    ----------
    t0 : str
        Start time window.
    t1 : str
        Stop time window
    
    Returns
    -------
    out: list of dict
        Program list.
    
    """
    if not isinstance(t0, int) and not isinstance(t0, np.int64):
        t0 = to_timestamp(t0)
    if not isinstance(t1, int) and not isinstance(t1, np.int64):
        t1 = to_timestamp(t1)
    url = __db_url__ + "programs.json?from=%d&upto=%d" %(t0, t1)
    try:    
        d = get_json(url)["programs"]
    except urllib2.HTTPError as ex:
        print(ex.read())
        return []
    return d

def get_program_list_for_day(day):
    """Get list of performed experimental programs for experimental day.

       This function returns a list of dictionaries for programs
       performed on a given day. The date should be
       specified as '2016-03-10'.
    Parameters
    ----------
    day : str
        Day of interest.
    
    Returns
    -------
    out: list of dict
        Program list.
    """
    day = day.strip()
    return get_program_list(day + " 00:00:01", day + " 23:59:59")

def get_program_id(t0, dt = 2, useCache = True):
    """
        Return program id (aka shot number) from ArchiveDB.

        Parameters
        ----------
        t0 : int or str
            The start time or the first t0 trigger of the run, or the experiment start.
        dt : int
            Time window in minutes. This is used to form a time window
            for searching.

        Returns
        -------
        out: str
            Program id as string. This has the form '20160310.002'
    """

    if useCache and t0 in __t02id__:
        return __t02id__[t0]
    t1 = t0 + dt*60*1000000000
    progs = archivedb.get_program_list(t0, t1)
    if not(len(progs)): # empty
        return ""
    ind = 0
    for i in range(1, len(progs)):
        if abs(progs[i]["trigger"]["0"][0] - t0) <  abs(progs[ind]["trigger"]["0"][0] - t0):
            ind = i
    now = int(time.time())
    if useCache and (now - t0*1e-9)>300:
        __t02id__[t0] = progs[ind]["id"]
    return progs[ind]["id"]

def get_mdsplus_shot(t0):
    """
        Return MDSPlus shot number closest to the trigger t0.

        This uses webservice query, which may fail.

        Parameters
        ----------
        t0 : int or str
            The first t0 trigger of the run, or the experiment start.
            This can be as nanosecond time stamp or time '2016-03-10 10:00:02.12132'.

        Returns
        -------
        out: str
    """
    if not isinstance(t0, int) and not isinstance(t0, np.int64):
        t0 = to_timestamp(t0)
    address =  __mdsplus_query_+"%d" %t0
    try:    
        d = get_json(address)
        return d["shotnum"]
    except urllib2.HTTPError as ex:
        print(ex.read())
        return 0

def get_program_t0(id, useCache=True):
    """
        Return program start as nanosecond time stamp. 
    
        This function returns the first t0 of the program. 
        The function takes  program id (aka shot number) in form
        '20160308.03'.

        Parameters
        ----------
        id : str

        Returns
        -------
        out: int
    """
    if useCache and id in __id2t0__:
        return __id2t0__[id]
    day = id[:4]+"-"+id[4:6]+"-"+id[6:8]
    progs = archivedb.get_program_list_for_day(day)
    now = int(time.time())
    for prog in progs:
        if prog["id"] == id:
            t0 = prog["trigger"]["0"][0]
            if useCache and (now - t0*1e-9)>300:
                __id2t0__[id] = t0
            return t0
    return 0

def get_program_from_to(id, useCache=True):
    """
        Return program from and upto time stamps.
    
        The function takes  program id (aka shot number) in form
        '20160308.03'.

        Parameters
        ----------
        id : str

        Returns
        -------
        from, upto: int, int
    """
    if useCache and id in __id2fromto__:
        return __id2fromto__[id][0], __id2fromto__[id][1]
    day = id[:4]+"-"+id[4:6]+"-"+id[6:8]
    progs = archivedb.get_program_list_for_day(day)
    now = int(time.time())
    for prog in progs:
        if prog["id"] == id:
            t0 = prog["upto"]
            if useCache and (now - t0*1e-9)>300:
                __id2fromto__[id] = [prog["from"], prog["upto"]]
            return prog["from"], prog["upto"]
    return 0,1

def get_base_address(signal_name):
    """
        Helper function to deal with URLs.
    """
    if not isinstance(signal_name, ("".__class__, u"".__class__)):
        signal_name = "/".join(str(c) for c in signal_name)
    if signal_name.find(__db_url__)==-1:
        baseaddress = __db_url__+ str(signal_name)
    else:
        baseaddress = signal_name
    baseaddress = baseaddress.replace(" ","%20")
    return baseaddress

def get_version_info(stream_name, version):
    """
        Get version information.

        Parameters
        ----------
        stream_name: str
            Name of data of parlog stream.
        version: str or int
            Version number to get info for. Either a string like 'V1' or int as 1,2.
            If None, info for all versions is returned

        Returns
        -------
        out: dict
            Version information.
    """
    address = get_base_address(stream_name)
    if address[-1] == "/":
        address = address[:-1]
    if isinstance(version, ("".__class__, u"".__class__)):
        address = address + "/"+version+"/_versions.json"
    elif isinstance(version, int) or isinstance(version, np.int64):
        address = address + "/V"+str(version)+"/_versions.json"
    elif version is None:
        address = address + "/_versions.json"
    return get_json(address)

def get_last_version(stream_name, time_from, time_to):
    """
        Find the latest available version in the given time interval.

        Parameters
        ----------
        stream_name: str
            Name of data of parlog stream.
        time_from: int or string
            The time from which to search.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time until which to search.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.

        Returns
        -------
        out: str
            Version number like V1 or None.
    """
    address = get_base_address(stream_name)
    if address[-1] == "/":
        address = address[:-1]
    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if time_to is None:
        time_to = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
    address = address + "/_versions.json?from=" + str(time_from) + "&upto=" + str(time_to)
    d = get_json(address)["versionInfo"]
    if len(d):
        return "V"+str(d[0]["number"])
    else:
        return None

def get_last_version_for_program(stream_name, id):
    """
        Find the latest available version for a program.

        Parameters
        ----------
        stream_name: str
            Name of data of parlog stream.
        id : str
            Program id, of the form "20160310.002".

        Returns
        -------
        out: str
            Version number like V1.
    """
    time_from, time_to = get_program_from_to(id)
    if time_from ==0:
        return None
    return get_last_version(stream_name, time_from, time_to)

def get_time_intervals(signal_name, time_from, time_to, maxIntervals = None):
    """
        Return time intervals where data samples exist.

        Note that the intervals are ordered from the newest to the oldest.
        
        Parameters
        ----------
        signal_name : string or list
            The signal address either as full URL, a string  with signal path, 
            or a list with signals path. 
            'Database/View/Project/StreamGroup /Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup", "Stream" ,
                "Channel#" , "Channel Name"]
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        maxIntervals : int, optional
            Return at most that number of intervals. If None all possible
            intervals are returned.

        Returns
        -------
        get_time_intervals : nd.array
            2d array of all time intervals where the data samples exist.
            Each element of the list contains a 2-integer element list 
            indicating the upper and lower end of the interval.
        See Also
        --------
        get_signal : Get the signal details within a given time stamp interval.
        get_latest_time_interval : Return latest valid time stamp interval where data samples
        (signal) exist.
        
        Examples
        --------
        >>> archivedb.get_all_time_intervals(
                'Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/...',
                '2015-05-29 00:00:00','2015-06-30 00:00:00')
        array([[1435601915999999999, 1435601931999600000], ....
               [1433159240999999999, 1433159256999600000]])
    """
    
    baseaddress = get_base_address(signal_name)
    
    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if time_to is None:
        time_to = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
    
    i = 0
    address =  baseaddress+'?filterstart='+ str(time_from)+ '&filterstop='+ str(time_to)
    timestamp_intervals = []
    while True:
        try:
            d = get_json(address)
        except urllib2.HTTPError as ex:
            msg = ex.read()
            raise RuntimeError(msg)
        try:
            urllinks = d['_links']['children']
         
            for link in urllinks:
                url = link["href"]
                pos1 = url.find('from=')+5
                pos2 = pos1+19
                t0 = int(url[pos1:pos2])
                pos1 = url.find('upto=')+5
                pos2 = pos1+19
                t1 = int(url[pos1:pos2])
                timestamp_intervals.append([t0, t1])
                if maxIntervals is not None and len(timestamp_intervals) >= maxIntervals:
                    break
        except Exception:
            pass

        # to avoid retrieving next pages
        if maxIntervals is not None and len(timestamp_intervals) >= maxIntervals:
            break

        try:
            address = d['_links']['next']['href']
            del d
            continue
        except Exception:
            del d
            break
    return np.array(timestamp_intervals)     


def get_latest_time_interval(signal_name):
    
    """
        Return latest valid time stamp interval where data samples exists
        
        Parameters
        ----------
        signal_name : string or list
            The signal address either as full URL, a string  with signal path, 
            or a list with signals path. 
            'Database/View/Project/StreamGroup /Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup", "Stream" ,
                "Channel#" , "Channel Name"]

        Returns
        -------
        get_latest_time_interval : ndarray
            The last time intervals with data samples.
            Element of the list contains a 2-integers, the lower and upper bound of the interval.

        See Also
        --------
        get_signal : Get the signal details within a given timestamp interval.
        get_time_intervals : Return valid timestamp intervals where data samples exist.
        Examples
        --------
        >>> archivedb.get_latest_time_interval(
                'ArchiveDB/raw/W7X/CoDaStationDesc.84/DataModuleDesc.187_DATASTREAM/0/...')
        array([1436356790231205506, 1436356800000000000])
    """
    res =  get_time_intervals(signal_name, "2010-01-01 00:00:01",
                              None, maxIntervals = 1)
    return res.squeeze()

def get_time_intervals_for_program(signal_name, id):
    """
        Find time intervals with data for given program id.

        Parameters
        ----------
        signal_name : string or list
            The signal address either as full URL, a string  with signal path, 
            or a list with signals path. 
            'Database/View/Project/StreamGroup /Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup", "Stream" ,
                "Channel#" , "Channel Name"]
        id : str
            Program id, of the form "20160310.002".
        Returns
        -------
        out : nd.array
            2d array of all time intervals where the data samples exist.
    """
    time_from, time_to = get_program_from_to(id)
    if time_from ==0:
        return [[]]
    return get_time_intervals(signal_name, time_from, time_to)
    
def get_signal(signal_name, time_from, time_to, useCache = True):
    """
        Get the signal data within a given timestamp interval.

        It is recommended to call this function with valid interval
        boundaries. In such a way the caching can be used effectively.
        
        Parameters
        ----------
        signal_name : string or list
            The signal address either as full URL, a string  with signal path, 
            or a list with signals path. 
            'Database/View/Project/StreamGroup /Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup", "Stream" ,
                "Channel#" , "Channel Name"]
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        Returns
        -------
        times, values : two arrays
            Time stamp and value arrays.
        See Also
        --------
        get_latest_time_interval : Return latest valid time stamp interval
        get_time_intervals : Return valid time stamp intervals
        
        get_all_signal : Get ALL the signal details (Values) in an array of a
        given timestamp intervals.
        
        Examples
        --------
        >>> archivedb.get_signal(
                'ArchiveDB/raw/W7X/CoDaStationDesc.115/DataModuleDesc.255_DATASTREAM/0/...',
                1439391817693999999,1439391817694000000)
        (array([1439391817694000000]), array([-0.4439773]))

    """
    baseaddress = get_base_address(signal_name)
    if baseaddress.startswith("http://archive-webapi.ipp-hgw.mpg.de/"):
        signal_name = baseaddress[len("http://"):]
    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
        
    times = np.array([time_from, time_to])
    filename = get_cache_name(signal_name, times) + ".npz"

    if useCache==True and os.path.exists(filename):
        return get_signal_from_cache(signal_name, times)
 
    address = baseaddress +"/_signal.json"    
    address =  address+'?from='+ str(int(time_from))+ '&upto='+ str(int(time_to))
    try:
        d = get_json(address)
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)
    x = np.array(d['dimensions'])
    y = np.array(d['values'])
    if useCache:
        write_signal_to_cache(signal_name, x, y, time_from, time_to)
    del d
    return x,y
  

def get_signal_multiinterval(signal_name, intervals, useCache=True):
    """
        Get the signal data  for multiple intervals.

        Note: intervals are inverted, to be usable together with get_time_intervals.

        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database / View / Project/ 
            StreamGroup / Stream_DATASTREAM / Channel # /Channel Name '
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        intervals : 2 dimensional nd.array
            A 2d array of time intervals.

        Returns
        -------
        times, values : ndarray,ndarray            
            Time stamp and value arrays. All intervals are joined together.

        Examples
        --------
        >>> x=archivedb.get_all_time_intervals('Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM
            /0/HBCm_U30_01','2015-05-29 00:00:00','2015-06-30 00:00:00')
            
    """        
    app_time = []
    app_signal = []
    
    for interval in intervals[::-1]:
        if not len(interval):
            continue
        temp_time,temp_signal = archivedb.get_signal(signal_name, interval[0], interval[1], useCache=useCache)
        app_time.extend(temp_time.tolist())
        app_signal.extend(temp_signal.tolist())
        del temp_time, temp_signal
    return np.array(app_time),np.array(app_signal)

def get_signal_for_program(signal_name, id, useCache=True, correctTime=True, useLastVersion=False):
    """
        Return data for given program.

        Parameters
        ----------
        signal_name : string or list
            The signal address either as full URL, a string  with signal path, 
            or a list with signals path. 
            'Database/View/Project/StreamGroup /Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup", "Stream" ,
                "Channel#" , "Channel Name"]
        id : str
            Program id, of the form "20160310.002".
        correctTime: bool, optional
            If true the time vector is aligned to the plasma start, i.e. (t - t0)*1e-9 - 60.0

        Returns
        -------
        times, values : ndarray,ndarray            
            Time stamp and value arrays. All intervals are joined together.
        
    """
    if useLastVersion:
        #assume version is not in the signal name
        p = signal_name.rfind("/", 0, signal_name.rfind("/"))
        stream = signal_name[:p]
        channel = signal_name[p:]
        version = get_last_version_for_program(stream, id)
        if version is not None:
            signal_name = stream + "/" + version + channel

    intervals = get_time_intervals_for_program(signal_name, id)
    if len(intervals) == 1 and len(intervals[0]) == 0:
        return np.array([]), np.array([])
    t, s = get_signal_multiinterval(signal_name, intervals, useCache=useCache)
    if correctTime:
        t0 = get_program_t0(id)
        t = (t - t0)*1e-9 - 60.0
    return t, s


def get_signal_box(signal_name, time_from, time_to, channels, useCache=True):
    """
        Get signal box for the given interval.

        Signal box is a set of several channels.
        
        Parameters
        ----------
        signal_name : string        
            The signal address of the form 'View/Project/StreamGroup/Stream/'
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        channels: list
            The channels for which data is required
            Of the form [3,4]
        Returns
        -------
        times, values : ndarray,ndarray            
            Time stamp and value arrays. 

        Examples
        --------
        >>> archivedb.get_signal_box(
                "ArchiveDB/raw/W7X/CoDaStationDesc.114/DataModuleDesc.256_DATASTREAM/",
                1407920479822000000, 1407920481823000000, [139,140,141])
            (array([1407920479822000000, 1407920481823000000]),
             array([[ 0.0049    ,  0.0049    ],
                    [ 0.96000004,  0.96000004],
                    [ 0.5       ,  0.5       ]]))

    """
    
    baseaddress = get_base_address(signal_name)
    if baseaddress.startswith("http://archive-webapi.ipp-hgw.mpg.de/"):
        signal_name = baseaddress[len("http://"):]+ "/"
    address = baseaddress + "/_signal.json"

    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)

    times = np.array([time_from, time_to])
    channels_to_download = []
    result = {}
    dim = None

    #try to retrieve data from cache
    for x in channels:
        indvch_address = str(signal_name) + str(x)
        filename = get_cache_name(indvch_address, times) + ".npz"
        if useCache==True and os.path.exists(filename):
            dim, val = get_signal_from_cache(indvch_address, times)
            result[x] = val
        else:
            channels_to_download.append(x)   
    
    # for channels that are not in the cache do download
    if channels_to_download:
        address =  address+'?from='+ str(int(time_from))+ '&upto='+ str(int(time_to)) + "&channels="+",".join([str(c) for c in channels_to_download])
        try:
            d = get_json(address)
        except urllib2.HTTPError as ex:
            msg = ex.read()
            raise RuntimeError(msg)
        dim = np.array(d['dimensions'])
        val = np.array(d['values'])
        
        for i in range(len(channels_to_download)):
            x = channels_to_download[i]
            indvch_address = str(signal_name)+ str(x)
            result[x] = val[i]
            write_signal_to_cache(indvch_address, dim, val[i], time_from, time_to)
       
    all_data = []
    for c in channels:
        all_data.append(result[c])
    all_data = np.array(all_data)
    del result
    return dim, all_data


def get_signal_box_multiinterval(signal_name, intervals, channels, useCache=True):
    """
        Read signal box for multiple time intervals.

        Signal box is a set of several channels.

        Note: intervals are inverted, to be usable together with get_time_intervals.

        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database / View / Project/ 
            StreamGroup / Stream_DATASTREAM / Channel # /Channel Name '
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        intervals : 2 dimensional nd.array
            A 2d array of time intervals.
        channels: list
            The channels for which data is required
            Of the form [3,4]

        Returns
        -------
        times, values : ndarray,ndarray            
            Time stamp and value arrays. All intervals are joined together.
    """
    baseaddress = get_base_address(signal_name)
    if baseaddress.startswith("http://archive-webapi.ipp-hgw.mpg.de/"):
        signal_name = baseaddress[len("http://"):]        
        
    app_time = []
    app_signal = []

    for i in range(len(channels)):
        app_signal.append([])
       
    for interval in intervals[::-1]:
        if not len(interval):
            continue
        temp_time,temp_signal = get_signal_box(baseaddress, interval[0], interval[1], channels, useCache=useCache)
        app_time.extend(temp_time.tolist())
        for i in range(len(channels)):
            app_signal[i].extend(temp_signal[i].tolist())
        del temp_time, temp_signal
    return np.array(app_time),np.array(app_signal)

def get_signal_box_for_program(signal_name, id, channels, useCache=True, correctTime=True, useLastVersion=False):
    """
        Return data for given program.

        Signal box is a set of several channels.
        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database / View / Project/ 
            StreamGroup / Stream_DATASTREAM / Channel # /Channel Name '
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        id : str
            Program id, of the form "20160310.002".
        channels: list
            The channels for which data is required
            Of the form [3,4]
        correctTime: bool, optional
            If true the time vector is aligned to the plasma start, i.e. (t - t0)*1e-9 - 60.0

        Returns
        -------
        times, values : ndarray,ndarray            
            Time stamp and value arrays. All intervals are joined together.
    """
    if useLastVersion:
        version = get_last_version_for_program(signal_name, id)
        if version is not None:
            if signal_name[-1] == "/":
                signal_name = signal_name[:-1]
            signal_name = signal_name + "/" + version
    # The real difficulty is to get the available intervals.
    # One needs a full channel name to ask for intervals.
    # Try to get one.
    baseaddress = get_base_address(signal_name)
    address = baseaddress + "/{:d}".format(channels[0])
    try:
        d = get_json(address)
        sname = d["_links"]["children"][0]["href"].split("?")[0]
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)
    intervals = get_time_intervals_for_program(sname, id)
    if len(intervals) == 1 and len(intervals[0]) == 0:
        return np.array([]), np.array([])
    t, s = get_signal_box_multiinterval(signal_name, intervals, channels, useCache=useCache)
    if correctTime:
        t0 = get_program_t0(id)
        t = (t - t0)*1e-9 - 60.0
    return t, s

def get_image_png(signal_name, time_from, time_to, useCache = True,
        timeout=1, retries=10):
    """
        Returns the last image within a given interval.

        The image is returned as a 2d array, but for data transfer
        png is used.
        
        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database/View/Project/
            StreamGroup/Stream_DATASTREAM/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        Returns
        -------
        time stamp, image : int, 2d numpy array
            Time stamp of the image and 2d data array.
        Examples
        --------
        >>> archivedb.get_image_png('Test/raw/Data4SoftwareTest/Hdf5.video.Test2
            /pixelfly_DATASTREAM',1421229109876999999,1421229109967000000)
            
            [[   0    0    0 ...,    0    0 1285]
             [   0    0    0 ...,    0    0 1285]
             [   0    0    0 ...,    0    0 1285]
             ..., 
             [   0    0    0 ...,    0    0    0]
             [   0    0    0 ...,    0    0    0]
             [   0    0    0 ...,    0    0    0]]
    """
    baseaddress = get_base_address(signal_name)
    if baseaddress.startswith("http://archive-webapi.ipp-hgw.mpg.de/"):
        signal_name = baseaddress[len("http://"):]

    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
        
    times = np.array([time_from, time_to])
    filename = get_cache_name(signal_name, times) + ".npz"
    if useCache==True and os.path.exists(filename):
        return get_image_from_cache(signal_name, times)
    
    baseaddress = baseaddress + "/_signal.png"
    address =  baseaddress+'?from='+ str(int(time_from))+ '&upto='+ str(int(time_to))
    
    request = urllib2.Request(address)
    response = None
    for i in range(retries):
        try:
            response = urllib2.urlopen(request, timeout=timeout)
        except socket.timeout:
            continue
    if response is None:
        raise RuntimeError("failed to urlopen")

    img = Image.open(BytesIO(response.read()))
   
    pixelarray = np.array(img.getdata()).reshape(img.size[1], img.size[0])

    y = int(response.info()["W7X-Timestamp"])
    if useCache:
        write_image_to_cache(signal_name, y, pixelarray, time_from, time_to)
    
    del request, response
    return y, pixelarray

def get_image_png_multiple(signal_name, time_from, time_to, useCache = True):
    """
        Returns the all images within a given interval.

        The images are returned as a 3d array, but for data transfer
        png is used. This uses multiple calls to get_image_png.

        Note: the images are sorted from the oldest to the newest.
        
        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database/View/Project/
            StreamGroup/Stream_DATASTREAM/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        Returns
        -------
        time stamp, image : array, 3d numpy array
            Time stamps of the images and 3d data array.
        Examples
        --------
        >>> archivedb.get_image_png('Test/raw/Data4SoftwareTest/Hdf5.video.Test2
            /pixelfly_DATASTREAM',1421229109876999999,1421229109967000000)
            
            [[   0    0    0 ...,    0    0 1285]
             [   0    0    0 ...,    0    0 1285]
             [   0    0    0 ...,    0    0 1285]
             ..., 
             [   0    0    0 ...,    0    0    0]
             [   0    0    0 ...,    0    0    0]
             [   0    0    0 ...,    0    0    0]]
    """
    times = []
    data = []
    intervals = get_time_intervals(signal_name, time_from, time_to)
    for interval in intervals:
        x,y = get_image_png(signal_name, interval[0], interval[1], useCache=useCache)
        times.append(x)
        data.append(y)
    return np.array(times), np.array(data)

def get_image_json(signal_name, time_from, time_to, correct_negatives = True):
    """
        Returns images a sequence of images as a 3d array.

        Data are returned as a 3d array, but the data transfer is done with JSON.

        Note: at the moment no cache implementation for this function.

        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database/View/Project/
            StreamGroup/Stream_DATASTREAM/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup",
            "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        correct_negatives : bool, optional
            If to correct negative values. This is sometimes required, because
            the actual unsigned data are represented as data with sign.
        Returns
        -------
        time stamp, image : int, 2d numpy array
            Time stamps of the images and 3d data array.
        
        See Also
        --------
        get_image_png : Returns the last image as a 2d array.
        
        Examples
        --------
        >>> archivedb.get_image_json('Test/raw/Data4SoftwareTest/Hdf5.video.Test2
            /pixelfly_DATASTREAM',1421229109876999999,1421229109967000000)
            
            [[[   0    0    0 ...,    0    0 1285]
              [   0    0    0 ...,    0    0 1285]
              [   0    0    0 ...,    0    0 1285]
              ..., 

              [   0    0    0 ...,    0    0    0]
              [   0    0    0 ...,    0    0    0]
              [   0    0    0 ...,    0    0    0]]]
            
            [1421229109877000000, 1421229109887000000, 1421229109897000000,
            1421229109907000000, 1421229109917000000, 
            1421229109927000000, 1421229109937000000, 1421229109947000000,
            1421229109957000000, 1421229109967000000]
    """
    baseaddress = get_base_address(signal_name) + "/_signal.json"
    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
    address =  baseaddress+'?from='+ str(int(time_from))+ '&upto='+ str(int(time_to))
    
    try:
        image_data = get_json(address)
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)

    data = np.array(image_data["values"])
    if correct_negatives:
        data[data<0] = data[data<0] + 32768 + 32768 
    return np.array(image_data["dimensions"]), data

def write_signal(signal_name, timestamp_list, data_list):
    """    
        Writes signal to database.
        
        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup",
                "Stream" , "Channel#" , "Channel Name"]
        timestamp_list : list
            The list of time in nanoseconds, 
            of the form: [1435601915999999999,1435601915999999998,1435601915999999992]
        data_list : list
            The signal data, of the form: [1, 2, 3]
        Returns
        -------
        out : boolean
            True if the data is successfully created.
 
        Examples
        --------
        >>> my_timestamp = int(round(time.time() * 1000)) * 1000000
            timestamps=[my_timestamp, my_timestamp+1, my_timestamp+2,my_timestamp+3,my_timestamp+4]
            
            archivedb.write_signal(
                    'Test/raw/W7XAnalysis/webapi-tests/pythonTest1_DATASTREAM/',
                    timestamps, [11,2,3,4,15])
            
            {"values": [11, 2, 3, 4, 15], "dimensions": [1437061349923000000, 1437061349923000001,
                1437061349923000002, 1437061349923000003, 1437061349923000004]}
    """
    
    if not isinstance(timestamp_list, list):
        timestamp_list = timestamp_list.tolist()
    if not isinstance(data_list, list):
        data_list = data_list.tolist()   
    url = get_base_address(signal_name) 
 
    msg = json.dumps({'values': data_list, 'dimensions': timestamp_list})
    request = urllib2.Request(url, data = msg.encode('utf-8'), headers={"Content-type" : "application/json"})

    try:
        response = urllib2.urlopen(request)
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)
    
    return (response.code == 201)



def get_parameters_box(signal_name, time_from, time_to):
    """
        Get the parameter box  within a given time interval.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream_DATASTREAM/Channel#/Channel Name'
        time_from: int or string
            The time from which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
        time_to: int or string
            The time upto which valid time intervals is to be searched.
            Of the form 'YYYY-MM-DD HH:MM:SS.%f', or nanosecond time stamp.
            If None is supplied, current time is used.
        Returns
        -------
        out : dict
            Set of parameters.
        
        Examples
        --------
        >>> archivedb.get_parameters_box(
                "ArchiveDB/raw/W7X/CoDaStationDesc.14823/DataModuleDesc.14833_PARLOG",
                1438604887140000000,1438604887240000000)
        
            {u'dimensions': [1438604887140000000, 1439539695409999999],
             u'label': u'parms',
             u'values': [{
                ...
               u'sysState': 5,
               u'sysText': u'',
               u'sysVersionInfo': u'',
               u'writeTimeVector': 1}]}
    """
    
    baseaddress = get_base_address(signal_name)
    if baseaddress.startswith("http://archive-webapi.ipp-hgw.mpg.de/"):
        signal_name = baseaddress[len("http://"):]
    baseaddress = baseaddress + "/_signal.json"
    if not isinstance(time_from, int) and not isinstance(time_from, np.int64):
        time_from = archivedb.to_timestamp(time_from)
    if not isinstance(time_to, int) and not isinstance(time_to, np.int64):
        time_to = archivedb.to_timestamp(time_to)
    
    address = baseaddress+'?from='+ str(int(time_from))+ '&upto='+ str(int(time_to))
    try:
        d = get_json(address)
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)
    return d

def get_parameters_box_for_program(signal_name, id):
    """
        Get the parameter box  within the time interval of a program.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream_DATASTREAM/Channel#/Channel Name'
        id : str
            Program id, of the form "20160310.002".
        Returns
        -------
        out : dict
            Set of parameters.
        
        Examples
        --------
        >>> archivedb.get_parameters_box(
                "ArchiveDB/raw/W7X/CoDaStationDesc.14823/DataModuleDesc.14833_PARLOG",
                "20171017.008")
        
            {u'dimensions': [1438604887140000000, 1439539695409999999],
             u'label': u'parms',
             u'values': [{
                ...
               u'sysState': 5,
               u'sysText': u'',
               u'sysVersionInfo': u'',
               u'writeTimeVector': 1}]}
    """
    time_from, time_to = get_program_from_to(id)
    if time_from ==0:
        return {}
    return get_parameters_box(signal_name, time_from, time_to)

def write_parameters(signal_name, data_dict, time_interval):
    """    
        Writes parameter log (_PARLOG) for into database.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 
            'Database/View/Project/StreamGroup/Stream_DATASTREAM/Channel#/Channel Name'
        data_dict : dict
            Parametes as a Python dictionary of the form:
            Example:  
                    { 'chanDescs': {
                                        '[0]':{
                                            'name' : 'Alpha', 
                                            'active' : 1, 
                                            'physicalQuantity' : { 'type' : 'X'}
                                        },
                                        '[1]' : 
                                        {
                                            'name' : 'Beta', 
                                            'active' : 1, 
                                            'physicalQuantity' : { 'type' : 'X'}
                                        }
                                                 },
                       'powerLevel' : 100 } 

        time_interval : 2 values list
            Time intervals when parameters are valid: ns timestamps

        Returns
        -------
        out : boolean 
            True if the data is successfully created.

        Examples
        --------
        >>> name = 'Test/raw/W7XAnalysis/webapi-tests/pythonTest1_PARLOG/'
        >>> t = int(round(time.time() * 1000)) * 1000000
        >>> archivedb.write_parameters('Test',name, para_data, [t, -1])
    """
    url = get_base_address(signal_name) 

    msg = {"label": "parms", "values":[data_dict,], "dimensions": time_interval}
    
    para_msg = json.dumps(msg)
    
    #print para_msg
    
    request = urllib2.Request(url, data = para_msg.encode('utf-8'),
                              headers={"Content-type" : "application/json"})
    try:
        response = urllib2.urlopen(request)
    except urllib2.HTTPError as ex:
        msg = ex.read()
        raise RuntimeError(msg)
    return (response.code == 201)

def write_parameters_from_ini_file(streamGroup, fileName, time_interval,
                                  add = None, addChannels = None):
    """    
        Read parameters from an ini file and write them to a parameter log (_PARLOG).

        Creates a settings dictionary form an ini file. An ini file looks like:
        [Section 1]
        par 1 = value1
        par 2 = value2
        [Section 2]
        .....
        Sections that must be present are: Info and Channel X, where X are channel numbers.
        The Info sections must have entries name and serial number, they are used to form
        the full signal name.

        Parameters are written to: streamGroup/name_serialNumber_PARLOG.

        
        Parameters
        ----------
        streamGroup : string        
            The adress of the stream group to use.
            'Database/View / Project /StreamGroup ', e.g.
            "Test/raw/W7X/thoms_test"
        fileName : string
            Name of the input ini file.
        time_interval : 2 values list
            Time intervals when parameters are valid: ns timestamps
        add : dictionary
            An optional dictionary of additional settings to add.
        addChannels: dictionary
            An optional dictionary of additional settings that is applied for
            every channel entry in chanDescs.

        Returns
        -------
        out : boolean 
            True if the data is successfully created.

        Examples
        --------
        >>> archivedb.write_parameters_from_ini_file('Test/raw/W7X/thoms_test',
                                        "ADQ14DC_SPD-04385.ini", [1, -1])
    """
    config = configparser.ConfigParser()
    config.read(fileName)
    settings = {}
    settings["chanDescs"] = collections.OrderedDict() # important! DB expects proper order
    channels = []
    for s in config.sections():
        if s[:7] == "Channel":                
            n = int(s.split(" ")[-1])
            ref = "[%d]" %n
            settings["chanDescs"][ref] = {}
            settings["chanDescs"][ref]["name"] = "channel_%d" %n
            settings["chanDescs"][ref]["active"] = 1
            for opt in config.options(s):
                settings["chanDescs"][ref][opt] = config.get(s, opt)
            if addChannels is not None:
                for k in addChannels.keys():
                    settings["chanDescs"][ref][k] = addChannels[k]
            continue
        settings[s] = {}
        for opt in config.options(s):
            settings[s][opt] = config.get(s, opt)
    if add is not None:
        for k in add.keys():
            settings[k] = add[k]
    if streamGroup[-1] == '/':
        streamGroup = streamGroup[:-1]
    name = "%s/%s_%s_PARLOG" %(streamGroup, config.get("Info", "name"),
                                            config.get("Info", "serial number"))
    return write_parameters(name, settings, time_interval)
        

def write_image_to_database(signal_name,timestamp, im, parameters):
    
    """    
        Writes image, time stamp and the parameters (PARLOG) to the database.
        
        Parameters
        ----------
        signal_name : string or list
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream_DATASTREAM/Channel#/Channel Name'
            or ["Database" , "View" , "Project" , "StreamGroup",
                    "Stream_DATASTREAM" , "Channel #" , "Channel Name"]
        timestamp : int
            Time in nanoseconds when the image was taken
        im : nd.array
            A 2d array of the image taken.
        parameters : dict
            
            Example : d =  {"projectName":"test data", "cameraType":"test data",
                            "height" : 287,
                            "width" : 380,
                            "bitDepth" : 10,
                            "dataBoxSize" : 1,
                            "datatype" : "short",
                            "unsigned" : 1}
        Returns
        -------
        out : boolean
            True if the image is successfully written.
        
        Examples
        --------
        >>> im = PIL.Image.open("w7x.jpg")
            width, height = im.size
            image = np.array(im.getdata()).reshape(height, width, 3).sum(axis=-1)
            d =  {"projectName":"test data", "cameraType":"test data",
            "height" : 287,
            "width" : 380,
            "bitDepth" : 10,
            "dataBoxSize" : 1,
            "datatype" : "short",
            "unsigned" : 1}
        
            archivedb.write_image_to_database("Test/raw/W7XAnalysis/HDF5_import_test
            /",archivedb.to_timestamp("2015-08-03 12:30:00"), image, d)
            
    """
    
    baseaddress = get_base_address(signal_name)
    randomname = str(random.randint(0,100))+".h5"
    height,width = im.shape
    image = np.zeros((height, width, 1), int)
    image[...,0] = im
    
    h5file = tables.open_file(randomname, "a", driver="H5FD_CORE", driver_core_backing_store=0)
    resp = 0
    try:
        h5file.create_group("/", "data")
        h5file.create_array("/data", "timestamps",
        obj=np.array([timestamp]))

        h5file.create_array("/data", "pixelfly", obj=image.astype(np.uint16))
        h5file.create_group("/", "pixelfly")

        h5file.create_group("/pixelfly", "parms")

        i = 0
        pk = list(parameters.keys())
        pv = list(parameters.values())
        while i != len(pk):
            try:
                h5file.create_array("/pixelfly/parms", pk[i], obj=pv[i])
            except Exception:
                h5file.create_array("/pixelfly/parms", pk[i], obj=pv[i].encode())
            i = i+1

        h5file.create_group("/pixelfly/parms", "chanDescs")
        for x in h5file.root.pixelfly.parms:
            if isinstance(x, tables.Group):
                continue
            del x.attrs.TITLE
            del x.attrs.FLAVOR
            del x.attrs.CLASS
            del x.attrs.VERSION

        h5file.create_group("/pixelfly/parms/chanDescs", "[0]")
        try:
            h5file.create_array("/pixelfly/parms/chanDescs/[0]", "name", obj="image")
        except Exception:
            h5file.create_array("/pixelfly/parms/chanDescs/[0]", "name", obj="image".encode())
        h5file.create_array("/pixelfly/parms/chanDescs/[0]", "active", obj=np.int32(1))
        for x in h5file.get_node("/pixelfly/parms/chanDescs/[0]"):
            del x.attrs.TITLE
            del x.attrs.FLAVOR
            del x.attrs.CLASS
            del x.attrs.VERSION

        path_query = '?dataPath=/data/pixelfly' + '&timePath=/data/timestamps'+ '&parameterPath=/pixelfly/parms'  
        buf = h5file.get_file_image()
        req = urllib2.Request(url=baseaddress + path_query, data=buf, 
                              headers={'Content-Type': 'application/x-hdf',
                                       'Content-Length': '%d' % len(buf)})
        try:
            resp = urllib2.urlopen(req).code
        except urllib2.HTTPError as ex:
            msg = ex.read()
            raise RuntimeError(msg)
    finally:
        h5file.close()   
    return resp == 201

def write_signal_to_cache(signal_name, times, signalvalues, time_from, time_to):
    """
        Write times and values to cache.

        Saves a file with the given signal name and timestamps.
        Of the form : ".archivedb/cache/Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/...
        /1433159241000000000_1435601931999600000.npz" 
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream/Channel#/Channel Name'
        times : nd.array
            Array of time stamps.
            Of the form [1437399819544000000, 1437399819544000001, 1437399819544000002]
        values : nd.array
            Values. 
            Of the form : [-0.01590027, -0.01480957, -0.01590271, -0.01480957]
            
        time_from : int
            Lower limit of the timestamp in the file. Used for naming the file.
            
        time_to : int
            Upper limit of the timestamp in the file. Used for naming the file.

        Returns
        -------
        None
            
        See Also
        --------
        get_signal_from_cache : Get the times and the values from cache.
        
        Examples
        --------
        >>> archivedb.write_signal_to_cache(
                'Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/HBCm_U30_01',
            [1437399819544000000, 1437399819544000001, 1437399819544000002],
            [-0.01590027, -0.01480957, -0.01590271], 
            1437399819544000000, 1437399819544000002)
    """
    
    filename = archivedb.get_cache_name(signal_name, np.array([time_from, time_to]))
    
    directory = filename[0 : filename.rfind("/")]

    if not os.path.exists(directory):
        os.makedirs(directory)
    
    np.savez(filename, timestamp = times, signal = signalvalues)
    
def get_cache_name(signal_name, times):
    """
        Returns cache filename.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream/Channel#/Channel Name'
        times : nd.array
            The array with 2 elements, the start and end points (in nanoseconds) of the time
            window.
            Of the form : [1437399819544000000, 1437399819544000002]
        Returns
        -------
        out : str
            Filename of the .npz file which stores time stamps and signal values.
        See Also
        --------
        write_signal_to_cache : Write timestamps and signal values to cache.
        
        Examples
        --------
        >>> archivedb.get_cache_name(
                'Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/HBCm_U30_01',
            [1437399819544000000, 1437399819544000001])
            "./.archivedb/cache/Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/HBCm_U30_01
                /1437399819544000000_1437399819544000001"
    """
    
    filename = __cache_path__ + signal_name + "/" + str(times.min()) + "_" + str(times.max())
    return filename

def get_signal_from_cache(signal_name, times):
    """
        Get time stamps and values in a given time interval from cache.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database/View/Project/StreamGroup/
            Stream/Channel#/Channel Name'
        times : nd.array
            Array with 2 elements, the start and end points (in nanoseconds)
            of the time window.
            Of the form [1437399819544000000, 1437399819544000001]

        Returns
        -------
        times, values : nd.array, nd.array
            Times and values arrays.
        
        See Also
        --------
        write_signal_to_cache : Write time stamps and values to cache.
        
        Examples
        --------
        >>> archivedb.get_signal_to_cache(
                'Test/raw/W7X/QSB_Bolometry/BoloSignal_DATASTREAM/0/HBCm_U30_01',
                [1437399819544000000, 1437399819544000001])
            
            array([1437399819544000000, 1437399819544000001, 1437399819544000001])
            array([-0.01590027, -0.01480957, -0.01590271])
    """
    
    filename = get_cache_name(signal_name, times) + ".npz"
    f = np.load(filename)
    return f["timestamp"], f["signal"]


def write_image_to_cache(signal_name, times, pixelarray, time_from, time_to):
    """
        Write the 2d array of an image and the corresponding time stamp to cache.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database / View / Project/ StreamGroup /
            Stream_DATASTREAM / Channel # /Channel Name '
        times : nd.array
            The timestamp when the image was taken, stored in a nd.array.
            Of the form : [1437399819544000000]
        pixelarray : nd.array
            The image in 2d numpy array
            Of the form :   [[   0    0    0 ...,    0    0 1285]
                             [   0    0    0 ...,    0    0 1285]
                             [   0    0    0 ...,    0    0 1285]
                             ..., 
                             [   0    0    0 ...,    0    0    0]
                             [   0    0    0 ...,    0    0    0]
                             [   0    0    0 ...,    0    0    0]]
            
        time_from : int
            Lower limit of the timestamp in the file. Used for naming the file.
            
        time_to : int
            Upper limit of the timestamp in the file. Used for naming the file.

        Returns
        -------
        write_image_to_cache : none
    """
    filename = archivedb.get_cache_name(signal_name, np.array([time_from, time_to]))
    directory = filename[0 : filename.rfind("/")]
    if not os.path.exists(directory):
        os.makedirs(directory)
    np.savez(filename, timestamp = times, image = pixelarray)

def get_image_from_cache(signal_name, times):
    """
        Get image(from a given time interval) in 2d array and the timestamp
        (when it was taken) from cache.
        
        Parameters
        ----------
        signal_name : string        
            The signal adress of the form 'Database / View / Project/ StreamGroup /
            Stream_DATASTREAM / Channel # /Channel Name '
        times : nd.array
            The start and end points (in nanoseconds) of the time window.
            Of the form [1437399819544000000, 1437399819544000001]

        Returns
        -------
        t, data : 2d nd.array, 1d nd.array
            Time of the image and data.
    """
    filename = archivedb.get_cache_name(signal_name, times) + ".npz"
    f = np.load(filename)
    return f["timestamp"], f["image"]




